using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Text;
using System.Threading;
using System.Windows.Forms;

namespace FileConvert
{
	public partial class MainForm : Form
	{
		ArrayList filesToConvert;
		long allFilesSize;
		MyEncoding inEnc;
		MyEncoding outEnc;
		byte lineEndings;
		Thread convertingThread = null;

		delegate void SetStatusDelegate(string text, int? progress);
		delegate void SetEnabledDelegate(bool enabled);

		public MainForm()
		{
			InitializeComponent();
		}

		private void MainForm_Load(object sender, EventArgs e)
		{
			DefaultEncoding.Text += " " + Encoding.Default.EncodingName + ", " + Encoding.Default.CodePage;
			FillEncodings();
			InputFilesList.Items.Add(new InputFileItem());
			CheckEnableConvert();
		}

		private void MainForm_DragEnter(object sender, DragEventArgs e)
		{
			if (e.Data.GetDataPresent("FileDrop"))
				e.Effect = DragDropEffects.Link;
		}

		private void MainForm_DragDrop(object sender, DragEventArgs e)
		{
			if (e.Data.GetDataPresent("FileDrop"))
			{
				for (int i = 0; i < InputFilesList.Items.Count; )
				{
					InputFileItem ifi = InputFilesList.Items[i] as InputFileItem;
					if (ifi.IsEmpty)
						InputFilesList.Items.Remove(ifi);
					else
						i++;
				}
				string[] files = (string[]) e.Data.GetData("FileDrop");
				foreach (string file in files)
				{
					InputFileItem newifi = new InputFileItem(file);
					if (InputEncoding.SelectedItem == null ||
						InputFilesList.Items.Count == 0)
					{
						try
						{
							// Auto-determine input file encoding
							FileStream stream = File.OpenRead(file);
							byte[] buffer = new byte[4];
							int codepage = 0;
							stream.Read(buffer, 0, 4);
							if (buffer[0] == 0xFF && buffer[1] == 0xFE && buffer[2] == 0x00 && buffer[3] == 0x00)
								codepage = 12000;   // UTF-32LE
							if (buffer[0] == 0x00 && buffer[1] == 0x00 && buffer[2] == 0xFE && buffer[3] == 0xFF)
								codepage = 12001;   // UTF-32BE
							if (buffer[0] == 0xEF && buffer[1] == 0xBB && buffer[2] == 0xBF)
								codepage = 65001;   // UTF-8
							if (buffer[0] == 0xFF && buffer[1] == 0xFE)
								codepage = 1200;   // UTF-16LE
							if (buffer[0] == 0xFE && buffer[1] == 0xFF)
								codepage = 1201;   // UTF-16BE
							stream.Close();
							foreach (MyEncoding enc in InputEncoding.Items)
							{
								if (enc.Encoding.CodePage == codepage)
									InputEncoding.SelectedItem = enc;
							}
						}
						catch
						{
						}
					}
					for (int i = 0; i < InputFilesList.Items.Count; )
					{
						InputFileItem ifi = InputFilesList.Items[i] as InputFileItem;
						if (ifi.Filename == newifi.Filename)
							InputFilesList.Items.Remove(ifi);
						else
							i++;
					}
					InputFilesList.Items.Add(newifi);
				}
				CheckEnableConvert();
			}
		}

		private void InputFilesList_KeyDown(object sender, KeyEventArgs e)
		{
			if (e.KeyCode == Keys.Delete && InputFilesList.SelectedIndices.Count > 0)
			{
				for (int index = 0; index < InputFilesList.Items.Count; )
				{
					if (InputFilesList.GetSelected(index))
						InputFilesList.Items.RemoveAt(index);
					else
						index++;
				}
				if (InputFilesList.Items.Count == 0)
				{
					InputFilesList.Items.Add(new InputFileItem());
					CheckEnableConvert();
				}
			}
			if (e.KeyCode == Keys.A && e.Control)
			{
				for (int index = 0; index < InputFilesList.Items.Count; index++)
					InputFilesList.SetSelected(index, true);
			}
		}

		private void InputEncoding_SelectedIndexChanged(object sender, EventArgs e)
		{
			CheckEnableConvert();
		}

		private void OutputDirectory_TextChanged(object sender, EventArgs e)
		{
			CheckEnableConvert();
		}

		private void OutputEncoding_SelectedIndexChanged(object sender, EventArgs e)
		{
			if (OutputEncoding.SelectedItem != null)
			{
				OutputUseBom.Enabled = ((MyEncoding) OutputEncoding.SelectedItem).IsUTF;
				if (!OutputUseBom.Enabled)
					OutputUseBom.Checked = false;
			}
			CheckEnableConvert();
		}

		private void AsInput_Click(object sender, EventArgs e)
		{
			if (InputFilesList.Items.Count > 0)
			{
				InputFileItem item = InputFilesList.Items[0] as InputFileItem;
				if (item != null && !item.IsEmpty)
				{
					OutputDirectory.Text = Path.GetDirectoryName(item.Filename);
				}
			}
		}

		private void Browse_Click(object sender, EventArgs e)
		{
			folderBrowserDialog1.SelectedPath = OutputDirectory.Text;
			DialogResult res = folderBrowserDialog1.ShowDialog();
			if (res == DialogResult.OK)
				OutputDirectory.Text = folderBrowserDialog1.SelectedPath;
		}

		private void ShowAllEncodings_CheckedChanged(object sender, EventArgs e)
		{
			FillEncodings();
		}

		private void Convert_Click(object sender, EventArgs e)
		{
			StatusLabel.Text = "";
			
			try
			{
				SetEnabled(false);

				// Create target directory
				if (!Directory.Exists(OutputDirectory.Text))
				{
					try
					{
						Directory.CreateDirectory(OutputDirectory.Text);
					}
					catch (Exception ex)
					{
						StatusLabel.Text = "Error creating output directory";
						throw ex;
					}
				}

				// Process all input files
				filesToConvert = new ArrayList();
				allFilesSize = 0;
				try
				{
					foreach (object obj in InputFilesList.Items)
					{
						InputFileItem item = obj as InputFileItem;
						if (item == null) continue;
						RememberConvertFile(item.Filename, Path.GetFileName(item.Filename));
					}

					inEnc = InputEncoding.SelectedItem as MyEncoding;
					outEnc = OutputEncoding.SelectedItem as MyEncoding;
					lineEndings = 0;
					if (LineEndingsDos.Checked) lineEndings = 1;
					if (LineEndingsUnix.Checked) lineEndings = 2;

					convertingThread = new Thread(new ThreadStart(ConvertAllFiles));
					convertingThread.Priority = ThreadPriority.BelowNormal;
					convertingThread.Start();
				}
				catch (Exception ex)
				{
					StatusLabel.Text = "Error converting file";
					throw ex;
				}
			}
			catch
			{
				StatusProgress.Value = 0;
				SetEnabled(true);
			}
			finally
			{
			}
		}

		private void Cancel_Click(object sender, EventArgs e)
		{
			if (convertingThread != null && convertingThread.IsAlive)
			{
				convertingThread.Abort();
				SetEnabled(true);
			}
			else
				Close();
		}

		/// <summary>
		/// Convert a file
		/// </summary>
		/// <param name="infile">Full input file path</param>
		/// <param name="outfile">Output file relative to OutputDirectory</param>
		private void RememberConvertFile(string infile, string outfile)
		{
			if (Directory.Exists(infile))
			{
				// Item is a directory, process all files in it
				foreach (string pathfile in Directory.GetFiles(infile))
				{
					string file = Path.GetFileName(pathfile);
					RememberConvertFile(Path.Combine(infile, file), Path.Combine(outfile, file));
				}
			}
			else if (File.Exists(infile))
			{
				long size = new FileInfo(infile).Length;
				filesToConvert.Add(new RememberedFile(infile, outfile, size));
				allFilesSize += size;
			}
		}

		void SetStatus(string text, int? progress)
		{
			if (this.statusStrip1.InvokeRequired)
			{
				SetStatusDelegate d = new SetStatusDelegate(SetStatus);
				this.Invoke(d, new object[] { text, progress });
			}
			else
			{
				if (text != null) StatusLabel.Text = text;
				if (progress != null) StatusProgress.Value = (int) progress;
			}
		}

		void SetEnabled(bool enabled)
		{
			if (this.InputFilesList.InvokeRequired)
			{
				SetEnabledDelegate d = new SetEnabledDelegate(SetEnabled);
				this.Invoke(d, new object[] { enabled });
			}
			else
			{
				InputFilesList.Enabled = enabled;
				InputEncoding.Enabled = enabled;
				OutputDirectory.Enabled = enabled;
				AsInput.Enabled = enabled;
				Browse.Enabled = enabled;
				OutputEncoding.Enabled = enabled;
				OutputUseBom.Enabled = enabled;
				LineEndingsDos.Enabled = enabled;
				LineEndingsNoChange.Enabled = enabled;
				LineEndingsUnix.Enabled = enabled;
				ShowAllEncodings.Enabled = enabled;
				Convert.Enabled = enabled;

				if (enabled)
					Cancel.Text = "Exit";
				else
					Cancel.Text = "Cancel";
			}
		}

		private void ConvertAllFiles()
		{
			long bytesProcessed = 0;
			int filesProcessed = 0;

			SetStatus("Converting...", null);

			foreach (RememberedFile file in filesToConvert)
			{
				StreamReader reader;
				try
				{
					reader = new StreamReader(file.InputFile, inEnc.Encoding, true);
				}
				catch (Exception ex)
				{
					MessageBox.Show(ex.ToString(), "Error reading file", MessageBoxButtons.OK, MessageBoxIcon.Error);
					bytesProcessed += file.FileSize;
					continue;
				}

				string outfile = Path.Combine(OutputDirectory.Text, file.OutputFile);
				FileStream outstream;
				try
				{
					if (File.Exists(outfile))
					{
						DialogResult res = MessageBox.Show("File " + outfile + " already exists. Overwrite?", "Overwrite file", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Warning);
						if (res == DialogResult.Cancel) break;
						if (res == DialogResult.No) continue;
					}
					outstream = new FileStream(outfile, FileMode.Create);
				}
				catch (Exception ex)
				{
					MessageBox.Show(ex.ToString(), "Error writing file", MessageBoxButtons.OK, MessageBoxIcon.Error);
					bytesProcessed += file.FileSize;
					continue;
				}

				if (inEnc.IsUTF && OutputUseBom.Checked)
				{
					byte[] bytes = outEnc.BOM;
					outstream.Write(bytes, 0, bytes.Length);
				}

				int bytesRead = 0;
				while (!reader.EndOfStream)
				{
					int i = reader.Read();
					bytesRead++;   // This is only an approximation if input file is in multi-byte encoding
					if (lineEndings == 1)
					{
						while (i == '\r') i = reader.Read();
						if (i == '\n') outstream.WriteByte((byte) '\r');
					}
					else if (lineEndings == 2)
					{
						while (i == '\r') i = reader.Read();
					}
					byte[] bytes = outEnc.Encoding.GetBytes(new char[1] { (char) i });
					outstream.Write(bytes, 0, bytes.Length);

					if (bytesRead % 10000 == 0)
						SetStatus(null, (int) ((bytesProcessed + bytesRead) * 1000 / allFilesSize));
				}

				reader.Close();
				outstream.Close();
				
				bytesProcessed += file.FileSize;
				filesProcessed++;
				SetStatus(filesProcessed + "/" + filesToConvert.Count + " files converted",
					(int) (bytesProcessed * 1000 / allFilesSize));
			}

			SetStatus("Finished", 0);
			SetEnabled(true);
		}

		private void FillEncodings()
		{
			string inEnc = InputEncoding.Text;
			string outEnc = OutputEncoding.Text;

			InputEncoding.Items.Clear();
			OutputEncoding.Items.Clear();

			foreach (EncodingInfo ei in System.Text.Encoding.GetEncodings())
			{
				MyEncoding enc = new MyEncoding(ei.GetEncoding());
				if (ShowAllEncodings.Checked || enc.IsBasicSet)
				{
					InputEncoding.Items.Add(enc);
					OutputEncoding.Items.Add(enc);
				}
			}

			InputEncoding.Text = inEnc;
			OutputEncoding.Text = outEnc;
		}

		private void CheckEnableConvert()
		{
			if (InputFilesList.Items.Count > 0 &&
				!(InputFilesList.Items[0] as InputFileItem).IsEmpty &&
				InputEncoding.SelectedIndex != -1 &&
				OutputEncoding.SelectedIndex != -1 &&
				OutputDirectory.Text != "")
				Convert.Enabled = true;
			else
				Convert.Enabled = false;
		}

		private void linkLabel1_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
		{
			UpdateCheck.Check("http://beta.unclassified.de/projekte/fileconvert/update.xml");
		}
	}

	class InputFileItem : Object
	{
		string filename;

		public InputFileItem()
		{
			filename = "";
		}

		public InputFileItem(string str)
		{
			filename = str;
		}

		public bool IsEmpty
		{
			get { return filename == ""; }
		}
		
		public string Filename
		{
			get { return filename; }
		}

		public override string ToString()
		{
			if (IsEmpty) return "(Drag files to convert here)";
			
			string s = "";
			s = Path.GetFileName(filename);
			if (Path.GetFileName(Path.GetDirectoryName(filename)) != "")
				s = Path.GetFileName(Path.GetDirectoryName(filename)) +
					Path.DirectorySeparatorChar + s;
			if (Path.GetFileName(Path.GetDirectoryName(Path.GetDirectoryName(filename))) != "")
				s = "..." + Path.DirectorySeparatorChar + s;
			s = Path.GetPathRoot(filename) + s;
			return s;
		}
	}

	class RememberedFile : Object
	{
		public string InputFile;
		public string OutputFile;
		public long FileSize;

		public RememberedFile(string infile, string outfile, long size)
		{
			InputFile = infile;
			OutputFile = outfile;
			FileSize = size;
		}
	}
}