using System;
using System.Collections.Generic;
using System.Text;

namespace FileConvert
{
	class MyEncoding : Object
	{
		Encoding encoding;
		bool systemDefault = false;

		public MyEncoding()
		{
			this.encoding = Encoding.Default;
			this.systemDefault = true;
		}

		public MyEncoding(Encoding encoding)
		{
			this.encoding = encoding;
		}

		public bool IsBasicSet
		{
			get
			{
				return (encoding.CodePage >= 1200 && encoding.CodePage <= 1201 ||
					encoding.CodePage >= 1250 && encoding.CodePage <= 1258 ||
					encoding.CodePage >= 10000 && encoding.CodePage <= 10082 ||
					encoding.CodePage >= 28591 && encoding.CodePage <= 28605 ||
					encoding.CodePage == 65001);
			}
		}

		public bool IsUTF
		{
			get
			{
				return (encoding.CodePage >= 1200 && encoding.CodePage <= 1201 ||
					encoding.CodePage >= 12000 && encoding.CodePage <= 12001 ||
					encoding.CodePage == 65001);
			}
		}

		public byte[] BOM
		{
			get
			{
				switch (encoding.CodePage)
				{
					case 1200:   // UTF-16LE
						return Encoding.GetBytes(new char[1] { (char) 0xFEFF });
					case 1201:   // UTF-16BE
						return Encoding.GetBytes(new char[1] { (char) 0xFEFF });
					case 12000:   // UTF-32LE
						return Encoding.GetBytes(new char[1] { (char) 0xFEFF });
					case 12001:   // UTF-32BE
						return Encoding.GetBytes(new char[1] { (char) 0xFEFF });
					case 65001:   // UTF-8
						return Encoding.GetBytes(new char[1] { (char) 0xFEFF });
					default:
						return null;
				}
			}
		}

		public Encoding Encoding
		{
			get
			{
				return encoding;
			}
		}

		public override string ToString()
		{
			switch (encoding.CodePage)
			{
				case 1200:
					return "UTF-16LE (Unicode Little Endian)";
				case 1201:
					return "UTF-16BE (Unicode Big Endian)";
				case 12000:
					return "UTF-32LE (UCS-4 Little Endian)";
				case 12001:
					return "UTF-32BE (UCS-4 Big Endian)";
				case 65001:
					return "UTF-8";
				default:
					return encoding.EncodingName + ", " + encoding.CodePage;
			}
		}
	}
}
